@@ -1,6 +1,6 @@
 #############################################################################
 #
-# Copyright (c) 2009 Zope Corporation and Contributors.
+# Copyright (c) 2009 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -67,6 +67,22 @@ class Selenese(object):
     def waitForPageToLoad(self):
         self.selenium.wait_for_page_to_load(self.timeout * 1000)
 
+    def waitForPopUp(self, windowID=''):
+        self.selenium.wait_for_pop_up(windowID, self.timeout * 1000)
+
+    def selectPopUp(self, windowID='', wait=True):
+        if wait:
+            timeout = self.timeout * 1000
+        else:
+            timeout = 0
+        self.selenium.wait_for_pop_up(windowID, timeout)
+        self.selenium.select_pop_up(windowID)
+
+    @passthrough
+    def close(self):
+        self.selenium.deselectPopUp()
+        pass
+
     @passthrough
     def createCookie(self, nameAndValue, options):
         pass
@@ -75,6 +91,10 @@ class Selenese(object):
     def deleteCookie(self, name, options):
         pass
 
+    @passthrough
+    def deselectPopUp(self):
+        pass
+
     @passthrough
     def dragAndDropToObject(self, locatorSource, locatorDestination):
         pass
@@ -183,6 +203,9 @@ class Selenese(object):
     @assert_type('locator_pattern')
     @passthrough
     def getEval(self, script):
+        # Note: we use the locator_pattern because the script acts like a
+        # locator: we pass it through and Selenium returns a result we can
+        # compare with the pattern.
         pass
 
     @assert_type('locator_pattern')
@@ -195,6 +218,11 @@ class Selenese(object):
     def getValue(self, locator):
         pass
 
+    @assert_type(None)
+    @passthrough
+    def isAlertPresent(self):
+        pass
+
     @assert_type('locator')
     @passthrough
     def isElementPresent(self, locator):
@@ -215,6 +243,10 @@ class Selenese(object):
     def isTextPresent(self, pattern):
         pass
 
+    @passthrough
+    def getLocation(self):
+        pass
+
     # Assertions
 
     def assertTextPresent(self, pattern):
@@ -244,58 +276,86 @@ class Selenese(object):
     # Internal
 
     def __getattr__(self, name):
+        requested_name = name
+        def _getattr(name):
+            try:
+                return getattr(self, name)
+            except AttributeError:
+                raise AttributeError(requested_name)
+
+        # Generate a number of assertions that aren't implemented directly.
+        # Apply a fall-back chain that first implements waitFor* using
+        # corresponding assert* and maps verify* methods to corresponding
+        # assert*, then implements negative assert* by negating their positive
+        # counterparts. Generate only these three kinds of assertions. After
+        # each fall-back step, getattr() is called to make use of all directly
+        # implemented methods. If a method cannot be looked up in spite of the
+        # fall-back, raise an AttributeError that reports the method name
+        # actually used by client code.
         if name.startswith('waitFor'):
             name = name.replace('waitFor', 'assert', 1)
-            assertion = getattr(self, name)
+            assertion = _getattr(name)
             return (lambda *args, **kw:
                     self._waitFor(assertion, *args, **kw))
 
-        if 'Not' in name:
-            name = name.replace('Not', '', 1)
-            assertion = getattr(self, name)
-            return (lambda *args, **kw:
-                    self._negate(assertion, *args, **kw))
-
         if name.startswith('verify'):
             name = name.replace('verify', 'assert', 1)
-            return getattr(self, name)
+            return _getattr(name)
 
-        if name.startswith('assert'):
-            getter_name = name.replace('assert', 'get', 1)
-            getter = getattr(self, getter_name, None)
-            if getter is None:
-                getter_name = name.replace('assert', 'is', 1)
-                getter = getattr(self, getter_name, None)
-            if getter.assert_type == 'pattern':
-                return lambda pattern: self._assert_pattern(getter, pattern)
-            elif getter.assert_type == 'locator':
-                return lambda locator: self._assert(getter, locator)
-            elif getter.assert_type == 'locator_pattern':
-                return lambda locator, pattern: \
-                        self._assert_pattern(getter, pattern, locator)
+        if not name.startswith('assert'):
+            raise AttributeError(requested_name)
 
-        raise AttributeError(name)
+        if 'Not' in name:
+            name = name.replace('Not', '', 1)
+            assertion = _getattr(name)
+            return (lambda *args, **kw:
+                    self._negate(assertion, requested_name, *args, **kw))
 
-    def _assert(self, getter, *args, **kw):
-        if not getter(*args, **kw):
+        # Positive assertions are synthesised by looking up a getter method
+        # for a value, getting the value and evaluating it in an appropriate
+        # way. Getters may be named either get* or is*.
+        try:
+            getter = _getattr(name.replace('assert', 'get', 1))
+        except AttributeError:
+            getter = _getattr(name.replace('assert', 'is', 1))
+
+        if getter.assert_type == 'pattern':
+            return lambda pattern: self._assert_pattern(
+                getter, requested_name, pattern)
+        elif getter.assert_type == 'locator':
+            return lambda locator: self._assert(
+                getter, requested_name, locator)
+        elif getter.assert_type == 'locator_pattern':
+            return lambda locator, pattern: self._assert_pattern(
+                getter, requested_name, pattern, locator)
+        elif getter.assert_type is None:
+            return lambda: self._assert(getter, requested_name)
+        else:
+            raise ValueError('Unknown assert type %r for selenese method %r.'
+                             % (getter.assert_type, requested_name))
+
+    def _assert(self, getter, name, *args, **kw):
+        value = getter(*args, **kw)
+        if not value:
             raise self.failureException(
-                'not: %s(%r, %r)' % (getter.__name__, args, kw))
+                'Failed: %s -> %r' %
+                (self._call_repr(name, *args, **kw), value))
 
-    def _assert_pattern(self, getter, pattern, *args):
+    def _assert_pattern(self, getter, name, pattern, *args):
         result = getter(*args)
         if not selenese_pattern_equals(result, pattern):
             raise self.failureException(
-                '%r did not match expected %r'
-                % (result, pattern))
+                'Expected: %r, got: %r from %s' %
+                (pattern, result, self._call_repr(name, *args)))
 
-    def _negate(self, assertion, *args, **kw):
+    def _negate(self, assertion, name, *args, **kw):
         try:
             assertion(*args, **kw)
-        except self.failureException, e:
+        except self.failureException:
             return
         else:
-            raise self.failureException('not %r with %r' %
-                                        (assertion, (args, kw)))
+            raise self.failureException(
+                'Failed: ' + self._call_repr(name, *args, **kw))
 
     def _waitFor(self, assertion, *args, **kw):
         start = time.time()
@@ -305,11 +365,17 @@ class Selenese(object):
             except self.failureException, e:
                 if time.time() - start > self.timeout:
                     raise self.failureException(
-                        'Timed out waiting for: %s' % e.args[0])
+                        'Timed out. %s' % e.args[0])
             else:
                 break
             time.sleep(0.1)
 
+    def _call_repr(self, name, *args, **kw):
+        return '%s(%s)' % (
+            name,
+            ', '.join(map(repr, args) +
+                      ['%s=%r' % item for item in sorted(kw.items())]))
+
 
 def match_glob(text, pattern):
     pattern = re.escape(pattern)
